import logging
from typing import List

import os
from twisted.internet.task import LoopingCall

from peek_platform.frontend.FrontendBuilderABC import FrontendBuilderABC
from peek_platform.frontend.FrontendOsCmd import runTsc

logger = logging.getLogger(__name__)


class NativescriptBuilder(FrontendBuilderABC):
    def __init__(self, frontendProjectDir: str, platformService: str,
                 jsonCfg, loadedPlugins: List):
        FrontendBuilderABC.__init__(self, frontendProjectDir, platformService,
                                    jsonCfg, loadedPlugins)

    def build(self) -> None:
        if not self._jsonCfg.feNativescriptBuildPrepareEnabled:
            logger.info("SKIPPING, Nativescript build prepare is disabled in config")
            return

        excludeRegexp = (
            r'.*[.]web[.]ts$',
            r'.*[.]mweb[.]ts$',
            r'.*[.]dweb[.]ts$',
            r'.*[.]web[.]html$',
            r'.*[.]mweb[.]html$',
            r'.*[.]dweb[.]html$',
            r'.*__pycache__.*',
            r'.*[.]py$'
        )

        self._dirSyncMap = list()

        feBuildDir = os.path.join(self._frontendProjectDir, 'build-ns')
        feSrcAppDir = os.path.join(self._frontendProjectDir, 'src', 'app')

        feAppDir = os.path.join(feBuildDir, 'app')
        feAssetsDir = os.path.join(feBuildDir, 'app', 'assets')

        self._moduleCompileRequired = False
        self._moduleCompileLoopingCall = None

        feModuleDirs = [
            (os.path.join(feAppDir, '@peek'), "moduleDir"),
        ]

        pluginDetails = self._loadPluginConfigs()

        # --------------------
        # Check if node_modules exists

        # if not os.path.exists(os.path.join(feBuildDir, 'node_modules')):
        #     raise NotADirectoryError("node_modules doesn't exist, ensure you've run "
        #                              "`npm install` in dir %s" % feBuildDir)

        # --------------------
        # Prepare the common frontend application

        self.fileSync.addSyncMapping(feSrcAppDir,
                                     os.path.join(feAppDir, 'app'),
                                     keepExtraDstJsAndMapFiles=True,
                                     excludeFilesRegex=excludeRegexp)

        # --------------------
        # Prepare the home and title bar configuration for the plugins
        self._writePluginHomeLinks(feAppDir, pluginDetails)
        self._writePluginTitleBarLinks(feAppDir, pluginDetails)

        # --------------------
        # Prepare the plugin lazy loaded part of the application
        self._writePluginRouteLazyLoads(feAppDir, pluginDetails)
        self._syncPluginFiles(feAppDir, pluginDetails, "appDir",
                              keepExtraDstJsAndMapFiles=True,
                              excludeFilesRegex=excludeRegexp)

        # --------------------
        # Prepare the plugin assets
        self._syncPluginFiles(feAssetsDir, pluginDetails, "assetDir",
                              excludeFilesRegex=excludeRegexp)

        # --------------------
        # Prepare the shared / global parts of the plugins

        self._writePluginRootModules(feAppDir, pluginDetails)
        self._writePluginRootServices(feAppDir, pluginDetails)

        # There are two
        for feModDir, jsonAttr, in feModuleDirs:
            # Link the shared code, this allows plugins
            # * to import code from each other.
            # * provide global services.
            self._syncPluginFiles(feModDir, pluginDetails, jsonAttr,
                                  keepExtraDstJsAndMapFiles=True,
                                  excludeFilesRegex=excludeRegexp)


        # Lastly, Allow the clients to override any frontend files they wish.
        self.fileSync.addSyncMapping(self._jsonCfg.feFrontendCustomisationsDir,
                                     feAppDir,
                                     parentMustExist=True,
                                     deleteExtraDstFiles=False,
                                     excludeFilesRegex=excludeRegexp)

        self.fileSync.syncFiles()

        if self._jsonCfg.feSyncFilesForDebugEnabled:
            logger.info("Starting frontend development file sync")
            self.fileSync.startFileSyncWatcher()

    def stopDebugWatchers(self):
        logger.info("Stoping frontend development file sync")
        self.fileSync.stopFileSyncWatcher()

    def _syncFileHook(self, fileName: str, contents: bytes) -> bytes:
        if fileName.endswith(".ts"):
            contents = contents.replace(b'@synerty/peek-util/index.web',
                                        b'@synerty/peek-util/index.ns')
            contents = contents.replace(b'@synerty/peek-util/index.mweb',
                                        b'@synerty/peek-util/index.ns')

            # replace imports that end with .web/.mweb with .ns
            # This will allow platform dependent typescript modules,
            # EG photo taking modules
            contents = contents.replace(b'.mweb";', b'.ns";')
            contents = contents.replace(b'.web";', b'.ns";')

            # Update the @peek import to use the /app path
            contents = contents.replace(b'from "@peek/', b'from "~/@peek/')

            # if b'@NgModule' in contents:
            #     return self._patchModule(fileName, contents)

            if b'@Component' in contents:
                return self._patchComponent(fileName, contents)

        return contents

    # def _patchModule(self, fileName: str, contents: bytes) -> bytes:
    #     newContents = b''
    #     for line in contents.splitlines(True):
    #
    #         newContents += line
    #
    #     return newContents

    def _patchComponent(self, fileName: str, contents: bytes) -> bytes:
        """ Patch Component
        
        Apply patches to the WEB file to convert it to the NativeScript version

        :param fileName: The name of the file
        :param contents: The original contents of the file
        :return: The new contents of the file
        """
        inComponentHeader = False

        newContents = b''
        for line in contents.splitlines(True):
            if line.startswith(b"@Component"):
                inComponentHeader = True

            elif line.startswith(b"export"):
                inComponentHeader = False

            elif inComponentHeader:
                line = (line
                        .replace(b'.mweb;"', b'.ns";')
                        .replace(b'.mweb.html', b'.ns.html')
                        .replace(b'.mweb.css', b'.ns.css')
                        .replace(b'.mweb.scss', b'.ns.scss')
                        .replace(b'.web;"', b'.ns";')
                        .replace(b'.web.html', b'.ns.html')
                        .replace(b'.web.css', b'.ns.css')
                        .replace(b'.web.scss', b'.ns.scss')
                        .replace(b"'./", b"'")
                        )

            newContents += line

        return newContents



"""
 *
 *  Copyright Synerty Pty Ltd 2013
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""
import logging
import os
import sys
import tarfile
import urllib.error
import urllib.parse
import urllib.request
from abc import ABCMeta
from typing import Optional

from pytmpdir.Directory import Directory
from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from txhttputil.downloader.HttpFileDownloader import HttpFileDownloader
from txhttputil.util.DeferUtil import deferToThreadWrap

from peek_platform.WindowsPatch import isWindows
from peek_platform.util.PtyUtil import spawnPty, \
    logSpawnException

logger = logging.getLogger(__name__)

PEEK_PLATFORM_STAMP_FILE = 'stamp'
"""Peek Platform Stamp File, The file within the release that conatins the version"""


class PeekSwInstallManagerABC(metaclass=ABCMeta):
    """ Peek Software Install Manager ABC

    This class handles downloading the latest platform update from the server service
    installing it and then restarting this service.

    """

    def __init__(self):
        pass

    @classmethod
    def makeReleaseFileName(cls, version: str) -> str:
        """ Make Release File Name

        This method creates an absolute path/filename for a peek release given it's
        stamp/version

        :param version: The stamp/version of the release
        :return: The absolute filename of the peek-release tar file
        """

        from peek_platform import PeekPlatformConfig

        return os.path.join(
            PeekPlatformConfig.config.platformSoftwarePath,
            'peek-release-%s.tar.gz' % version)

    @classmethod
    def makePipArgs(cls, directory: Directory) -> [str]:
        """ Make PIP Args

        This method creates the install arg list for pip, it's used both when testing
        when a new platform release is uploaded and the install on each service.

        :param directory: The directory where the peek-release is extracted to
        :return: The list of arguments to pass to pip
        """
        # Create an array of the package paths

        absFilePaths = [f.realPath
                        for f in directory.files
                        if f.name.endswith(".tar.gz") or f.name.endswith(".whl")]

        # Create and return the pip args
        return ['install',  # Install the packages
                '--force-reinstall',  # Reinstall if they already exist
                '--no-cache-dir',  # Don't use the local pip cache
                '--no-index',  # Work offline, don't use pypi
                '--find-links', directory.path,
                # Look in the directory for dependencies
                ] + absFilePaths

    def notifyOfPlatformVersionUpdate(self, newVersion):
        self.installAndRestart(newVersion)

    @inlineCallbacks
    def update(self, targetVersion) -> Optional[str]:
        """ Update

        This method is called when this service detects that the peek server has a newer
        version of software than this service.

        :param targetVersion: The target version to update to
        :return: The version that was updated to, or None if it failed
        """
        logger.info("Updating to %s", targetVersion)

        from peek_platform import PeekPlatformConfig

        url = ('http://%(ip)s:%(port)s/peek_server.sw_install.platform.download?'
               ) % {"ip": PeekPlatformConfig.config.peekServerHost,
                    "port": PeekPlatformConfig.config.peekServerPort}

        args = {"name": PeekPlatformConfig.componentName}
        if targetVersion:
            args["version"] = str(targetVersion)

        url += urllib.parse.urlencode(args)

        try:
            file = yield HttpFileDownloader(url).run()

            if os.path.getsize(file.name) == 0:
                logger.warning("Peek server doesn't have any updates for %s, version %s",
                               PeekPlatformConfig.componentName, targetVersion)
                return

            yield self._installUpdate(targetVersion, file.name)

        except Exception as e:
            logger.exception(e)
            raise

        defer.returnValue(targetVersion)

    @inlineCallbacks
    def installAndRestart(self, targetVersion: str) -> None:
        newSoftwareTar = self.makeReleaseFileName(targetVersion)

        yield self._installUpdate(targetVersion, newSoftwareTar)

    @deferToThreadWrap
    def _installUpdate(self, targetVersion: str, fullTarPath: str) -> str:
        """ Install Update (Blocking)

        This method installs the packages in the latest peek-release.
        It then calls self.restartProcess to restart the service

        :param targetVersion: The version we should be updating to.
        :param fullTarPath: The path to the peek-release to install
        :return: The version that was installed, (from the file in the release)
        """

        from peek_platform import PeekPlatformConfig

        if not tarfile.is_tarfile(fullTarPath):
            raise Exception("Platform update download is not a tar file")

        directory = Directory()
        tarfile.open(fullTarPath).extractall(directory.path)
        directory.scan()

        stampFile = directory.getFile(name=PEEK_PLATFORM_STAMP_FILE)
        if not stampFile:
            raise Exception("Peek release %s doesn't contain version stamp file %s"
                            % (fullTarPath, PEEK_PLATFORM_STAMP_FILE))

        with stampFile.open() as f:
            stampVersion = f.read().strip()

        if stampVersion != targetVersion:
            raise Exception("Stamp file version %s doesn't match target version %s"
                            % (stampVersion, targetVersion))

        self._pipInstall(directory)

        PeekPlatformConfig.config.platformVersion = targetVersion

        # Call later, allow the server time to respond to the UI
        reactor.callLater(2.0, self.restartProcess)

        return targetVersion

    def _pipInstall(self, directory: Directory) -> None:
        """ Pip Install

        Runs the PIP install for the packages provided in the directory

        :param directory: The directory containing the
        :return: None

        """

        pipExec = os.path.join(os.path.dirname(sys.executable), "pip")

        pipArgs = [sys.executable, pipExec] + self.makePipArgs(directory)

        # The platform update is tested for dependencies when it's first uploaded
        # PIP has a bug, when you have updated packages for several dependent files
        #   and try to install them all at once, some of the packages don't update.
        pipArgs += ['--no-deps']

        pipArgs = ' '.join(pipArgs)

        try:
            spawnPty(pipArgs)
            logger.info("Peek package update complete.")

        except Exception as e:
            logSpawnException(e)

            # Update the detail of the exception and raise it
            e.message = "Failed to install packages from the new release."
            raise

    # @abstractmethod
    # def _stopCode(self) -> None:
    #     """ Stop Code
    #
    #     This method should stop the running code, mainly timers and processing queues.
    #     Possibly web servers that may call database updates.
    #
    #     """
    #
    # @abstractmethod
    # def _updateCode(self):
    #     """ Update Code
    #
    #     This method is called to update any data, such as a database migration (typically)
    #     """
    #
    # @abstractmethod
    # def _startCode(self):
    #     """ Start Code
    #
    #     This is called when the update fails, the service should start back up and run as
    #     it did before the update.
    #
    #     """

    # def _synlinkTo(self, componentName, home, newPath):
    #     symLink = os.path.join(home, componentName)
    #     try:
    #         os.remove(symLink)
    #     except:
    #         pass
    #     os.symlink(newPath, symLink)

    def restartProcess(self) -> None:
        """Restart Process

        Restarts the current program.

        Note: this function does not return.
        Any cleanup action (like saving data) must be done before calling this function.

        """
        python = sys.executable
        argv = list(sys.argv)
        argv.insert(0, "-u")

        def addExe(val):
            if not "run_peek_" in val:
                return val
            if isWindows and not val.lower().endswith(".exe"):
                return val + ".exe"
            return val

        argv = map(addExe, argv)
        os.execl(python, python, *argv)



import logging
import multiprocessing

from jsoncfg.value_mappers import require_string, require_integer

logger = logging.getLogger(__name__)


class PeekFileConfigWorkerMixin:
    @property
    def celeryBrokerUrl(self) -> str:
        # for BROKER_URL
        default = 'amqp://guest:guest@localhost:5672//'
        with self._cfg as c:
            return c.celery.brokerUrl(default, require_string)

    @property
    def celeryResultUrl(self) -> str:
        # for CELERYD_CONCURRENCY
        default = 'redis://localhost:6379/0'
        with self._cfg as c:
            return c.celery.resultUrl(default, require_string)

    @property
    def celeryWorkerCount(self) -> str:
        # for CELERYD_CONCURRENCY

        # By default, we assume a single server setup.
        # So leave half the CPU threads for the database
        default = int(multiprocessing.cpu_count() / 2)
        with self._cfg as c:
            return c.celery.workerCount(default, require_integer)

    @property
    def celeryTaskPrefetch(self) -> str:
        # for CELERYD_PREFETCH_MULTIPLIER

        default = 2

        with self._cfg as c:
            return c.celery.taskPrefetch(default, require_integer)



import logging
from abc import ABCMeta

from jsoncfg.value_mappers import require_string, require_dict
from peek_platform.file_config.PeekFileConfigABC import PeekFileConfigABC

logger = logging.getLogger(__name__)


class PeekFileConfigSqlAlchemyMixin:
    @property
    def dbConnectString(self):
        default = 'postgresql://peek:PASSWORD@127.0.0.1/peek'
        with self._cfg as c:
            return c.sqlalchemy.connectUrl(default, require_string)

    @property
    def dbEngineArgs(self):
        default = {
            'pool_size': 20,  # Number of connections to keep open
            'max_overflow': 50,  # Number that the pool size can exceed when required
            'pool_timeout': 60,  # Timeout for getting conn from pool
            'pool_recycle': 600  # Reconnect?? after 10 minutes
        }
        with self._cfg as c:
            return c.sqlalchemy.engineArgs(default, require_dict)



import logging
import typing
from datetime import datetime

import celery
from kombu import serialization

from peek_platform.file_config.PeekFileConfigWorkerMixin import PeekFileConfigWorkerMixin
from vortex.DeferUtil import noMainThread
from vortex.Payload import Payload

logger = logging.getLogger(__name__)


def vortexDumps(arg: typing.Tuple) -> str:
    noMainThread()
    # startTime = datetime.utcnow()
    try:
        return Payload(tuples=[arg])._toJson()
    except Exception as e:
        logger.exception(e)
        raise
    # finally:
    #     logger.debug("vortexDumps took %s", (datetime.utcnow() - startTime))


def vortexLoads(jsonStr: str) -> typing.Tuple:
    noMainThread()
    # startTime = datetime.utcnow()
    try:
        return Payload()._fromJson(jsonStr).tuples[0]
    except Exception as e:
        logger.exception(e)
        raise
    # finally:
    #     logger.debug("vortexLoads took %s", (datetime.utcnow() - startTime))


serialization.register(
    'vortex', vortexDumps, vortexLoads,
    content_type='application/x-vortex',
    content_encoding='utf-8',
)


def configureCeleryApp(app, workerConfig: PeekFileConfigWorkerMixin):
    # Optional configuration, see the application user guide.
    app.conf.update(
        BROKER_URL=workerConfig.celeryBrokerUrl,
        CELERY_RESULT_BACKEND=workerConfig.celeryResultUrl,

        # Leave the logging to us
        CELERYD_HIJACK_ROOT_LOGGER=False,

        # The time results will stay in redis before expiring.
        # I believe they are cleared when the results are obtained
        # from txcelery._DeferredTask
        CELERY_TASK_RESULT_EXPIRES=3600,

        # The number of tasks each worker will prefetch.
        CELERYD_PREFETCH_MULTIPLIER=workerConfig.celeryTaskPrefetch,

        # The number of workers to have at one time
        CELERYD_CONCURRENCY=workerConfig.celeryWorkerCount,

        # The maximum number or results to keep for the client
        # We could have backlog of 1000 results waiting for the client to pick up
        # This would be a mega performance issue.
        CELERY_MAX_CACHED_RESULTS=1000,  # Default is 100

        CELERY_TASK_SERIALIZER='vortex',
        # CELERY_ACCEPT_CONTENT=['vortex'],  # Ignore other content
        CELERY_ACCEPT_CONTENT=['pickle', 'json', 'msgpack', 'yaml', 'vortex'],
        CELERY_RESULT_SERIALIZER='vortex',
        CELERY_ENABLE_UTC=True,
    )



from peek_platform.file_config.PeekFileConfigABC import PeekFileConfigABC
from peek_platform.file_config.PeekFileConfigPlatformMixin import PeekFileConfigPlatformMixin

class _WorkerTaskConfigMixin(PeekFileConfigABC,
                        PeekFileConfigPlatformMixin):
    pass

@celery.signals.after_setup_logger.connect
def configureCeleryLogging(*args, **kwargs):
    # Fix the loading problems windows has
    from peek_plugin_base.PeekVortexUtil import peekWorkerName
    from peek_platform import PeekPlatformConfig
    PeekPlatformConfig.componentName = peekWorkerName
    config = _WorkerTaskConfigMixin()

    # Set default logging level
    logging.root.setLevel(config.loggingLevel)

    if config.loggingLevel != "DEBUG":
        for name in ("celery.worker.strategy", "celery.app.trace", "celery.worker.job"):
            logging.getLogger(name).setLevel(logging.WARNING)



import logging
from typing import List

import os

from peek_platform.frontend.FrontendBuilderABC import FrontendBuilderABC
from peek_platform.frontend.FrontendOsCmd import runNgBuild

logger = logging.getLogger(__name__)


class WebBuilder(FrontendBuilderABC):
    def __init__(self, frontendProjectDir: str, platformService: str,
                 jsonCfg, loadedPlugins: List):
        FrontendBuilderABC.__init__(self, frontendProjectDir, platformService,
                                    jsonCfg, loadedPlugins)

        self.isMobile = "mobile" in platformService
        self.isDesktop = "desktop" in platformService
        self.isAdmin = "admin" in platformService

    def build(self) -> None:
        if not self._jsonCfg.feWebBuildPrepareEnabled:
            logger.info("SKIPPING, Web build prepare is disabled in config")
            return

        excludeRegexp = (
            r'.*[.]ns[.]ts$',
            r'.*[.]ns[.]html$',
            r'.*__pycache__.*',
            r'.*[.]py$'
        )

        if self.isMobile:
            excludeRegexp += (
                r'.*[.]dweb[.]ts$',
                r'.*[.]dweb[.]html$',
            )

        elif self.isDesktop:
            excludeRegexp += (
                r'.*[.]mweb[.]ts$',
                r'.*[.]mweb[.]html$',
            )

        elif self.isAdmin:
            pass

        else:
            raise NotImplementedError("This is neither mobile or desktop web")

        self._dirSyncMap = list()

        feBuildDir = os.path.join(self._frontendProjectDir, 'build-web')
        feSrcAppDir = os.path.join(self._frontendProjectDir, 'src', 'app')

        feBuildSrcDir = os.path.join(feBuildDir, 'src')
        feBuildAssetsDir = os.path.join(feBuildDir, 'src', 'assets')

        feModuleDirs = [
            (os.path.join(feBuildSrcDir, '@peek'), "moduleDir"),
        ]

        pluginDetails = self._loadPluginConfigs()

        # --------------------
        # Check if node_modules exists

        if not os.path.exists(os.path.join(feBuildDir, 'node_modules')):
            raise NotADirectoryError("node_modules doesn't exist, ensure you've run "
                                     "`npm install` in dir %s" % feBuildDir)

        # --------------------
        # Prepare the common frontend application

        self.fileSync.addSyncMapping(feSrcAppDir, os.path.join(feBuildSrcDir, 'app'),
                                     excludeFilesRegex=excludeRegexp)

        # --------------------
        # Prepare the home and title bar configuration for the plugins
        self._writePluginHomeLinks(feBuildSrcDir, pluginDetails)
        self._writePluginTitleBarLinks(feBuildSrcDir, pluginDetails)

        # --------------------
        # Prepare the plugin lazy loaded part of the application
        self._writePluginRouteLazyLoads(feBuildSrcDir, pluginDetails)
        self._syncPluginFiles(feBuildSrcDir, pluginDetails, "appDir",
                              excludeFilesRegex=excludeRegexp)

        # --------------------
        # Prepare the plugin assets
        self._syncPluginFiles(feBuildAssetsDir, pluginDetails, "assetDir",
                              excludeFilesRegex=excludeRegexp)

        # --------------------
        # Prepare the shared / global parts of the plugins

        self._writePluginRootModules(feBuildSrcDir, pluginDetails)
        self._writePluginRootServices(feBuildSrcDir, pluginDetails)

        for feModDir, jsonAttr, in feModuleDirs:
            # Link the shared code, this allows plugins
            # * to import code from each other.
            # * provide global services.
            self._syncPluginFiles(feModDir, pluginDetails, jsonAttr,
                                  excludeFilesRegex=excludeRegexp)

        # Lastly, Allow the clients to override any frontend files they wish.
        self.fileSync.addSyncMapping(self._jsonCfg.feFrontendCustomisationsDir,
                                     feBuildSrcDir,
                                     parentMustExist=True,
                                     deleteExtraDstFiles=False,
                                     excludeFilesRegex=excludeRegexp)

        self.fileSync.syncFiles()

        if self._jsonCfg.feSyncFilesForDebugEnabled:
            logger.info("Starting frontend development file sync")
            self.fileSync.startFileSyncWatcher()

        if self._jsonCfg.feWebBuildEnabled:
            logger.info("Starting frontend web build")
            self._compileFrontend(feBuildDir)

    def _syncFileHook(self, fileName: str, contents: bytes) -> bytes:
        # replace imports that end with .dweb or .mweb to the appropriate
        # value
        # Otherwise just .web should be used if no replacing is required.

        if self.isMobile:
            contents = contents.replace(b'.dweb";', b'.mweb";')

        elif self.isDesktop:
            contents = contents.replace(b'.mweb";', b'.dweb";')

        elif self.isAdmin:
            pass

        else:
            raise NotImplementedError("This is neither mobile or desktop web")

        return contents

    def _compileFrontend(self, feBuildDir: str) -> None:
        """ Compile the frontend

        this runs `ng build`

        We need to use a pty otherwise webpack doesn't run.

        """

        hashFileName = os.path.join(feBuildDir, ".lastHash")

        if not self._recompileRequiredCheck(feBuildDir, hashFileName):
            logger.info("Frondend has not changed, recompile not required.")
            return

        logger.info("Rebuilding frontend distribution")

        try:
            runNgBuild(feBuildDir)

        except Exception as e:
            if os.path.exists(hashFileName):
                os.remove(hashFileName)

            # Update the detail of the exception and raise it
            e.message = "The angular frontend failed to build."
            raise



import logging
import os
import platform

logger = logging.getLogger(__name__)

isWindows = platform.system() is "Windows"


def createHardLink(src, dst):
    import ctypes
    flags = 1 if src is not None and os.path.isdir(src) else 0
    if not ctypes.windll.kernel32.CreateHardLinkA(dst, src, flags):
        raise OSError(ctypes.get_last_error())


def createSymbolicLink(src, dst, target_is_directory=False):
    """ Create Symbolic Link

    http://timgolden.me.uk/pywin32-docs/win32file__CreateSymbolicLink_meth.html
    """
    import win32file
    flags = 1 if target_is_directory else 0
    win32file.CreateSymbolicLink(dst, src, flags)


if isWindows:
    logger.info("Replacing os.link/os.symlink functions.")
    os.link = createHardLink
    os.symlink = createSymbolicLink



import gc
import logging
import sys
from typing import Type, Tuple, Optional

import os
from abc import ABCMeta, abstractmethod, abstractproperty
from collections import defaultdict
from importlib.util import find_spec
from jsoncfg.value_mappers import require_string, require_array
from twisted.internet.defer import inlineCallbacks

from peek_platform import PeekPlatformConfig
from peek_plugin_base.PluginCommonEntryHookABC import PluginCommonEntryHookABC
from peek_plugin_base.PluginPackageFileConfig import PluginPackageFileConfig
from vortex.PayloadIO import PayloadIO
from vortex.Tuple import removeTuplesForTupleNames, registeredTupleNames, \
    tupleForTupleName
from vortex.TupleAction import TupleGenericAction, TupleUpdateAction
from vortex.TupleSelector import TupleSelector
from vortex.rpc.RPC import _VortexRPCResultTuple, _VortexRPCArgTuple

logger = logging.getLogger(__name__)

# This doesn't do anything, but it makes sure it's imported before any plugins import it.
TupleSelector()
TupleUpdateAction()
TupleGenericAction()
_VortexRPCResultTuple()
_VortexRPCArgTuple()

corePlugins = [
    "peek_core_device"
]


class PluginLoaderABC(metaclass=ABCMeta):
    _instance = None

    def __new__(cls, *args, **kwargs):
        assert cls._instance is None, "PluginServerLoader is a singleton, don't construct it"
        cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self):
        self._loadedPlugins = {}

        self._vortexEndpointInstancesByPluginName = defaultdict(list)
        self._vortexTupleNamesByPluginName = defaultdict(list)

    @abstractproperty
    def _entryHookFuncName(self) -> str:
        """ Entry Hook Func Name.
        Protected property
        :return: EG  "peekServerEntryHook"

        """

    @abstractproperty
    def _entryHookClassType(self):
        """ Entry Hook Class Type
        Protected property
        :return: EG  PluginServerEntryHookABC

        """

    @abstractproperty
    def _platformServiceNames(self) -> [str]:
        """ Platform Service Name
        Protected property
        :return: one or more of "server", "worker", "agent", "client", "storage"

        """

    def pluginEntryHook(self, pluginName) -> Optional[PluginCommonEntryHookABC]:
        """ Plugin Entry Hook

        Returns the loaded plugin entry hook for the plugin name.

        :param pluginName: The name of the plugin to load

        :return: An instance of the plugin entry hook

        """
        return self._loadedPlugins.get(pluginName)

    @inlineCallbacks
    def loadPlugin(self, pluginName):
        try:
            self.unloadPlugin(pluginName)

            # Make note of the initial registrations for this plugin
            endpointInstancesBefore = set(PayloadIO().endpoints)
            tupleNamesBefore = set(registeredTupleNames())

            modSpec = find_spec(pluginName)
            if not modSpec:
                raise Exception("Failed to find package %s,"
                                " is the python package installed?" % pluginName)

            PluginPackage = modSpec.loader.load_module()
            pluginRootDir = os.path.dirname(PluginPackage.__file__)

            # Load up the plugin package info
            pluginPackageJson = PluginPackageFileConfig(pluginRootDir)
            pluginVersion = pluginPackageJson.config.plugin.version(require_string)
            pluginRequiresService = pluginPackageJson.config.requiresServices(
                require_array)

            # Make sure the service is required
            # Storage and Server are loaded at the same time, hence the intersection
            if not set(pluginRequiresService) & set(self._platformServiceNames):
                logger.debug("%s does not require %s, Skipping load",
                             pluginName, self._platformServiceNames)
                return

            # Get the entry hook class from the package
            entryHookGetter = getattr(PluginPackage, str(self._entryHookFuncName))
            EntryHookClass = entryHookGetter() if entryHookGetter else None

            if not EntryHookClass:
                logger.warning(
                    "Skipping load for %s, %s.%s is missing or returned None",
                    pluginName, pluginName, self._entryHookFuncName)
                return

            if not issubclass(EntryHookClass, self._entryHookClassType):
                raise Exception("%s load error, Excpected %s, received %s"
                                % (pluginName, self._entryHookClassType, EntryHookClass))

            ### Perform the loading of the plugin
            yield self._loadPluginThrows(pluginName, EntryHookClass,
                                         pluginRootDir, tuple(pluginRequiresService))

            # Make sure the version we have recorded is correct
            PeekPlatformConfig.config.setPluginVersion(pluginName, pluginVersion)

            # Make note of the final registrations for this plugin
            self._vortexEndpointInstancesByPluginName[pluginName] = list(
                set(PayloadIO().endpoints) - endpointInstancesBefore)

            self._vortexTupleNamesByPluginName[pluginName] = list(
                set(registeredTupleNames()) - tupleNamesBefore)

            self.sanityCheckServerPlugin(pluginName)

        except Exception as e:
            logger.error("Failed to load plugin %s", pluginName)
            logger.exception(e)

    @abstractmethod
    def _loadPluginThrows(self, pluginName: str,
                          EntryHookClass: Type[PluginCommonEntryHookABC],
                          pluginRootDir: str,
                          requiresService: Tuple[str, ...]) -> None:
        """ Load Plugin (May throw Exception)

        This method is called to perform the load of the module.

        :param pluginName: The name of the Peek App, eg "plugin_noop"
        :param EntryHookClass: The plugin entry hook class to construct.
        :param pluginRootDir: The directory of the plugin package,
         EG dirname(plugin_noop.__file__)

        """

    def unloadPlugin(self, pluginName: str):
        oldLoadedPlugin = self._loadedPlugins.get(pluginName)

        if not oldLoadedPlugin:
            return

        del oldLoadedPlugin

        # Remove the registered endpoints
        for endpoint in self._vortexEndpointInstancesByPluginName[pluginName]:
            PayloadIO().remove(endpoint)
        del self._vortexEndpointInstancesByPluginName[pluginName]

        # Remove the registered tuples
        removeTuplesForTupleNames(self._vortexTupleNamesByPluginName[pluginName])
        del self._vortexTupleNamesByPluginName[pluginName]

        self._unloadPluginPackage(pluginName)

    def listPlugins(self):
        def pluginTest(name):
            if not name.startswith("plugin_"):
                return False
            return os.path.isdir(os.path.join(self._pluginPath, name))

        plugins = os.listdir(self._pluginPath)
        plugins = list(filter(pluginTest, plugins))
        return plugins

    @inlineCallbacks
    def loadCorePlugins(self):
        for pluginName in corePlugins:
            yield self.loadPlugin(pluginName)

        # Start the Plugin
        for pluginName in corePlugins:
            if pluginName not in self._loadedPlugins:
                continue

            yield self._loadedPlugins[pluginName].start()

    def unloadCorePlugins(self):
        for pluginName in corePlugins:
            if pluginName in self._loadedPlugins:
                self.unloadPlugin(pluginName)

    @inlineCallbacks
    def loadOptionalPlugins(self):
        for pluginName in PeekPlatformConfig.config.pluginsEnabled:
            if pluginName.startswith("peek_core"):
                raise Exception("Core plugins can not be configured")
            yield self.loadPlugin(pluginName)

        # Start the Plugin
        for pluginName in PeekPlatformConfig.config.pluginsEnabled:
            if pluginName not in self._loadedPlugins:
                continue

            yield self._loadedPlugins[pluginName].start()

    def unloadOptionalPlugins(self):
        for pluginName in reversed(PeekPlatformConfig.config.pluginsEnabled):
            if pluginName in self._loadedPlugins:
                self.unloadPlugin(pluginName)

        remainingOptionalPlugins = list(filter(lambda n: not n.startswith("peek_core"),
                                               self._loadedPlugins))

        if remainingOptionalPlugins:
            logger.debug(remainingOptionalPlugins)
            raise Exception("Some plugins are still loaded")

    def _unloadPluginPackage(self, pluginName):

        oldLoadedPlugin = self._loadedPlugins.get(pluginName)

        # Stop and remove the Plugin
        del self._loadedPlugins[pluginName]

        try:
            oldLoadedPlugin.stop()
            oldLoadedPlugin.unload()

        except Exception as e:
            logger.error("An exception occured while unloading plugin %s,"
                         " unloading continues" % pluginName)
            logger.exception(e)

        # Unload the packages
        loadedSubmodules = [modName
                            for modName in list(sys.modules.keys())
                            if modName.startswith('%s.' % pluginName)]

        for modName in loadedSubmodules:
            del sys.modules[modName]

        if pluginName in sys.modules:
            del sys.modules[pluginName]

        gc.collect()

        # pypy doesn't have getrefcount
        # ("oldLoadedPlugin" in this method and the call to getrefcount) == 2 references
        if hasattr(sys, "getrefcount") and sys.getrefcount(oldLoadedPlugin) > 2:
            logger.warning("Old references to %s still exist, count = %s",
                           pluginName, sys.getrefcount(oldLoadedPlugin))

        del oldLoadedPlugin
        gc.collect()
        # Now there should be no references.

    def sanityCheckServerPlugin(self, pluginName):
        ''' Sanity Check Plugin

        This method ensures that all the things registed for this plugin are
        prefixed by it's pluginName, EG plugin_noop
        '''

        # All endpoint filters must have the 'plugin' : 'plugin_name' in them
        for endpoint in self._vortexEndpointInstancesByPluginName[pluginName]:
            filt = endpoint.filt
            if 'plugin' not in filt and filt['plugin'] != pluginName:
                raise Exception("Payload endpoint does not contan 'plugin':'%s'\n%s"
                                % (pluginName, filt))

        # all tuple names must start with their pluginName
        for tupleName in self._vortexTupleNamesByPluginName[pluginName]:
            TupleCls = tupleForTupleName(tupleName)
            if not tupleName.startswith(pluginName):
                raise Exception("Tuple name does not start with '%s', %s (%s)"
                                % (pluginName, tupleName, TupleCls.__name__))

    def notifyOfPluginVersionUpdate(self, pluginName, pluginVersion):
        logger.info("Received PLUGIN update for %s version %s", pluginName, pluginVersion)
        return self.loadPlugin(pluginName)
